/* dhcpd.c
 *
 * udhcp Server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <time.h>
#include <sys/time.h>

#include "debug.h"
#include "dhcpd.h"
#include "arpping.h"
#include "socket.h"
#include "options.h"
#include "files.h"
#include "leases.h"
#include "packet.h"
#include "serverpacket.h"
#include "pidfile.h"
#include "./gemtek_added_SRC/gemtek_jimmy_added_002.h"


/* globals */
struct dhcpOfferedAddr *leases;
struct server_config_t server_config;
static int signal_pipe[2];

/* Exit and cleanup */
static void exit_server(int retval)
{
	pidfile_delete(server_config.pidfile);
	CLOSE_LOG();
	exit(retval);
}


/* Signal handler */
static void signal_handler(int sig)
{
	if (send(signal_pipe[1], &sig, sizeof(sig), MSG_DONTWAIT) < 0) {
		LOG(LOG_ERR, "Could not send signal: %s", 
			strerror(errno));
	}
}


void		ProcessExistProcedure ( void )
{
	pid_t	pid = 0;
	int		status = 0 , i = 0;

RESTART_POINT:	
	if ( fork () == 0 ) {
		fprintf(stderr, "%s child process, pid = %d\n" , "dhcpd", getpid () );
	} else {
		fprintf(stderr, "%s parent process, pid = %d\n" , "dhcpd",getpid () );
		pid = wait ( &status );
		i = WEXITSTATUS ( status );
		sleep ( 10 );
		fprintf(stderr, "child's pid = %d , exit status = %d\n" , pid , i );	

		goto RESTART_POINT;
	}
} 

#ifdef COMBINED_BINARY	
int udhcpd_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{	
	fd_set rfds;
	struct timeval tv;
	int server_socket = -1;
	int bytes, retval;
	struct dhcpMessage packet;
	unsigned char *state;
	unsigned char *server_id, *requested, *hostname;
	u_int32_t server_id_align, requested_align;
	unsigned long timeout_end;
	struct option_set *option;
	struct dhcpOfferedAddr *lease;
	int pid_fd;
	int max_sock;
	int sig;
	SomeVariableDefinition_001

	OPEN_LOG("udhcpd");
	LOG(LOG_INFO, "udhcp server (v%s) started", VERSION);

	ProcessExistProcedure();

	memset(&server_config, 0, sizeof(struct server_config_t));
	
	if (argc < 2)
		read_config(DHCPD_CONF_FILE);
	else read_config(argv[1]);

	SomeExtraCodes_004
	pid_fd = pidfile_acquire(server_config.pidfile);
	pidfile_write_release(pid_fd);

	if ((option = find_option(server_config.options, DHCP_LEASE_TIME))) {
		memcpy(&server_config.lease, option->data + 2, 4);
		server_config.lease = ntohl(server_config.lease);
	}
	else server_config.lease = LEASE_TIME;
	
	leases = malloc(sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
	memset(leases, 0, sizeof(struct dhcpOfferedAddr) * server_config.max_leases);
	read_leases(server_config.lease_file);

	if (read_interface(server_config.interface, &server_config.ifindex,
			   &server_config.server, server_config.arp) < 0)
		exit_server(1);

#ifndef DEBUGGING
	pid_fd = pidfile_acquire(server_config.pidfile); /* hold lock during fork. */
	if (daemon(0, 0) == -1) {
		perror("fork");
		exit_server(1);
	}
	pidfile_write_release(pid_fd);
#endif


	socketpair(AF_UNIX, SOCK_STREAM, 0, signal_pipe);
	signal(SIGUSR1, signal_handler);
	signal(SIGTERM, signal_handler);

	timeout_end = time(0) + server_config.auto_time;
	SomeExtraCodes_001
	while(1) { /* loop until universe collapses */

		if (server_socket < 0)
			if ((server_socket = listen_socket(INADDR_ANY, SERVER_PORT, server_config.interface)) < 0) {
				LOG(LOG_ERR, "FATAL: couldn't create server socket, %s", strerror(errno));
				exit_server(0);
			}			

		FD_ZERO(&rfds);
		FD_SET(server_socket, &rfds);
		FD_SET(signal_pipe[0], &rfds);
		if (server_config.auto_time) {
			tv.tv_sec = timeout_end - time(0);
			tv.tv_usec = 0;
		}
		if (!server_config.auto_time || tv.tv_sec > 0) {
			max_sock = server_socket > signal_pipe[0] ? server_socket : signal_pipe[0];
			retval = select(max_sock + 1, &rfds, NULL, NULL, 
					server_config.auto_time ? &tv : NULL);
		} else retval = 0; /* If we already timed out, fall through */

		if (retval == 0) {
			write_leases();
			SomeExtraCodes_003
			timeout_end = time(0) + server_config.auto_time;
			continue;
		} else if (retval < 0 && errno != EINTR) {
			DEBUG(LOG_INFO, "error on select");
			continue;
		}
		
		if (FD_ISSET(signal_pipe[0], &rfds)) {
			if (read(signal_pipe[0], &sig, sizeof(sig)) < 0)
				continue; /* probably just EINTR */
			switch (sig) {
			case SIGUSR1:
				LOG(LOG_INFO, "Received a SIGUSR1");
				write_leases();
				SomeExtraCodes_003
				/* why not just reset the timeout, eh */
				timeout_end = time(0) + server_config.auto_time;
				continue;
			case SIGTERM:
				LOG(LOG_INFO, "Received a SIGTERM");
				write_leases();
				exit_server(0);
			}
		}

		if ((bytes = get_packet(&packet, server_socket)) < 0) { /* this waits for a packet - idle */
			if (bytes == -1 && errno != EINTR) {
				DEBUG(LOG_INFO, "error on read, %s, reopening socket", strerror(errno));
				close(server_socket);
				server_socket = -1;
			}
			continue;
		}

		if ((state = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL) {
			DEBUG(LOG_ERR, "couldn't get option from packet, ignoring");
			continue;
		}
		
		/* ADDME: look for a static lease */
		lease = find_lease_by_chaddr(packet.chaddr);
		switch (state[0]) {
		case DHCPDISCOVER:
			DEBUG(LOG_INFO,"received DISCOVER");
			
			if (sendOffer(&packet) < 0) {
				LOG(LOG_ERR, "send OFFER failed");
				SomeExtraCodes_005
			}
			break;			
 		case DHCPREQUEST:
			DEBUG(LOG_INFO, "received REQUEST");
			requested = get_option(&packet, DHCP_REQUESTED_IP);
			server_id = get_option(&packet, DHCP_SERVER_ID);
			hostname = get_option(&packet, DHCP_HOST_NAME);

			if (requested) memcpy(&requested_align, requested, 4);
			if (server_id) memcpy(&server_id_align, server_id, 4);
		
			if (lease) { /*ADDME: or static lease */
				if (server_id) {
					/* SELECTING State */
					DEBUG(LOG_INFO, "server_id = %08x", ntohl(server_id_align));
					if (server_id_align == server_config.server && requested && 
					    requested_align == lease->yiaddr) {
						sendACK(&packet, lease->yiaddr);
					}
					SomeExtraCodes_006
				} else {
					if (requested) {
						/* INIT-REBOOT State */
						if (lease->yiaddr == requested_align)
							sendACK(&packet, lease->yiaddr);
						else sendNAK(&packet);
					} else {
						/* RENEWING or REBINDING State */
						if (lease->yiaddr == packet.ciaddr)
							sendACK(&packet, lease->yiaddr);
						else {
							/* don't know what to do!!!! */
							sendNAK(&packet);
						}
					}						
				}
				if (hostname) {
					bytes = hostname[-1];
					if (bytes >= (int) sizeof(lease->hostname))
						bytes = sizeof(lease->hostname) - 1;
					strncpy(lease->hostname, hostname, bytes);
					lease->hostname[bytes] = '\0';
				} else
					lease->hostname[0] = '\0';
					
				//SomeExtraCodes_002
			/* what to do if we have no record of the client */
			} else if (server_id) {
				/* SELECTING State */
				SomeExtraCodes_007

			} else if (requested) {

				SomeExtraCodes_008
#ifndef	GemtekAddedSRC	// the original SRCs

				/* INIT-REBOOT State */
				if ((lease = find_lease_by_yiaddr(requested_align))) {
					if (lease_expired(lease)) {
						/* probably best if we drop this lease */
						memset(lease->chaddr, 0, 16);
					/* make some contention for this address */
					} else sendNAK(&packet);
				} else if (requested_align < server_config.start || 
					   requested_align > server_config.end) {
					sendNAK(&packet);
				} else {
					sendNAK(&packet);
				}
#else	// the following SRCs are added by Gemtek.
				if ( ( ntohl ( requested_align ) < ntohl ( server_config.start ) ) || 
							     ( ntohl ( requested_align ) > ntohl ( server_config.end ) ) ) { 
					DEBUG ( LOG_INFO , "The requested_align is in the wrong domain" ); 
					sendNAK ( &packet ); 
				} else if ( !add_lease ( packet.chaddr , requested_align , server_config.offer_time ) ) { 
					DEBUG ( LOG_INFO , "lease pool is full -- OFFER abandoned" ); 
					sendNAK ( &packet ); 
				} else { 
					DEBUG ( LOG_INFO , "lease pool is not full" ); 
					lease = find_lease_by_chaddr ( packet.chaddr ); 
					if ( lease ) { 
						DEBUG ( LOG_INFO , "No lease ... adding lease ... successful" ); 
						sendACK ( &packet , requested_align ); 
						if ( hostname ) { 
							bytes = hostname[-1]; 
							if (bytes >= (int) sizeof(lease->hostname)) 
								bytes = sizeof(lease->hostname) - 1; 
							strncpy(lease->hostname, hostname, bytes); 
							lease->hostname[bytes] = '\0'; 
						} else { 
							lease->hostname[0] = '\0'; 
						} 
					} else { 
						DEBUG ( LOG_INFO , "No lease ... adding lease ... failed" ); 
						sendNAK ( &packet ); 
					} 
				}
#endif	// end of ( #ifndef	GemtekAddedSRC )

			} else if (packet.ciaddr) {
				/* RENEWING or REBINDING State */

				SomeExtraCodes_010
#ifdef	GemtekAddedSRC
				
				if ( ( ntohl ( packet.ciaddr ) < ntohl ( server_config.start ) ) || 
						     ( ntohl ( packet.ciaddr ) > ntohl ( server_config.end ) ) ) { 
					DEBUG ( LOG_INFO , "The requested_align is in the wrong domain" ); 
					sendNAK ( &packet ); 
				} else if ( !add_lease ( packet.chaddr , packet.ciaddr , server_config.offer_time ) ) { 
					DEBUG ( LOG_INFO , "lease pool is full -- OFFER abandoned" ); 
					sendNAK(&packet); 
				} else { 
					DEBUG ( LOG_INFO , "lease pool is not full" ); 
					lease = find_lease_by_chaddr ( packet.chaddr ); 
					if ( lease ) { 
						DEBUG ( LOG_INFO , "No lease ... adding lease ... successful" ); 
						sendACK ( &packet , packet.ciaddr ); 
						if ( hostname ) { 
							bytes = hostname[-1]; 
							if (bytes >= (int) sizeof(lease->hostname)) 
								bytes = sizeof(lease->hostname) - 1; 
							strncpy(lease->hostname, hostname, bytes); 
							lease->hostname[bytes] = '\0'; 
						} else { 
							lease->hostname[0] = '\0'; 
						} 
					} else { 
						DEBUG ( LOG_INFO , "No lease ... adding lease ... failed" ); 
						sendNAK ( &packet ); 
					} 
				}
#else
				sendNAK(&packet);			// the original SRC
#endif	// end of GemtekAddedSRC

			}
			break;
		case DHCPDECLINE:
			DEBUG(LOG_INFO,"received DECLINE");
			if (lease) {
				memset(lease->chaddr, 0, 16);
				lease->expires = time(0) + server_config.decline_time;
			}			
			break;
		case DHCPRELEASE:
			DEBUG(LOG_INFO,"received RELEASE");
			if (lease) lease->expires = time(0);
			break;
		case DHCPINFORM:
			DEBUG(LOG_INFO,"received INFORM");
			send_inform(&packet);
			break;	
		default:
			LOG(LOG_WARNING, "unsupported DHCP message (%02x) -- ignoring", state[0]);
		}
	}

	return 0;
}

